// Copyright 2017 Kakao Corp. <http://www.kakaocorp.com>
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#pragma once

#include <chrono>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <string>
#include <vector>

namespace n2 {

class Data{
public:
    Data(const std::vector<float>& vec);
    inline const std::vector<float>& GetData() const { return data_; };
private:
    std::vector<float> data_;
};

float GetTimeDiff(const std::chrono::steady_clock::time_point& begin_t,
                  const std::chrono::steady_clock::time_point& end_t);

std::string GetCurrentDateTime();

} // namespace n2
